package se.sics.streamingclassifier

// ClassStatistics is the class which contains the parameters for the class feature distributions
// This is the data structure which you will update for each new observation and use to calculate the
// class-conditional probabilities. It encapsulates the statistics for a single class.
case class ClassStatistics(var classCount: Long, var classProbability: Double, var mean: Array[Mean], var variance: Array[Variance])

/**
 * A Naive Bayes classifier.
 * @param numClasses: The number of classes to use. Can be higher than the actual number of classes without problem.
 * @param numFeatures: The number of features of each observation.
 */
class NaiveBayesModel(numClasses: Int, numFeatures: Int) extends ClassificationModel(numFeatures, numClasses)
{
  // classStatistics is the central member of this model. It's an array indexed by the class label, where the elements
  // are ClassStatistics objects for that class. When a training sample is observed, the ClassStatistics object
  // corresponding to that class should be updated.
  var classStatistics: Array[ClassStatistics] =
  (0 until numClasses).map(i => {
    val count = 0
    val probability = 0
    val mean = Array.fill(numFeatures)(0.0)
    val variance = Array.fill(numFeatures)(0.0)
    new ClassStatistics(count, probability, mean, variance)}).toArray

  // totalCount keeps track of the total amount of training samples we've seen
  var totalCount: Long = 0

  /**
   * Returns the logarithm of the probability of x given a Gaussian with the given mean and variance
   * @param x: The value to calculate the log probability of.
   * @param mean: The mean of the Gaussian.
   * @param variance: The variance of the Gaussian.
   * @return The log probability of x for the given parameters.
   */
  def logGaussianPDF(x: Feature, mean: Mean, variance: Variance): Double = {
    // Since we're calculating the log of the gaussian, we can do certain simplifications. We don't need to use the
    // exponential function, and the numerator becomes
    val deviation = x - mean
    -0.5 * deviation * deviation / variance - 0.5 * Math.log(2 * Math.PI * variance)
  }

  /**
   * Updates the model parameters based on a new observation.
   * @param labeledSample: the new observation.
   */
  def update(labeledSample: LabeledSample): Unit = labeledSample match {
    case LabeledSample(label, features) =>
      // TODO: You need to update the statistics for the class corresponding to the given sample. Since we assume
      // TODO: the features are independent, each of them can be updated separately. Make sure you update the counters
      // TODO: as well.
      // TODO: The statistics are in the member classStatistics, which is an array of ClassStatstics object, see the
      // TODO: top of this file for details.

      // TODO: Write code here to update counters and classStatistics.
  }

  /**
   * Predicts the class for the given unlabeled sample.
   * @param unlabeledSample: The unlabeled sample to predict the class for.
   * @return The ClassLabel of the predicted class.
   */
  def predict(unlabeledSample: UnlabeledSample): ClassLabel = {
    // TODO: p should be an array with the class-wise log probabilities of observing the unlabeled sample.
    // TODO: You need to calculate the values for each of the elements of p. The method logGaussianPDF can be used
    // TODO: to get the log probability of a scalar, given a scalar mean and variance.
    val p: Array[Double] = Array.fill(numClasses)(Double.NegativeInfinity)

    // TODO: Write code here to update p.

    // The built in Array.max and Array.indexOf methods seem to have problems
    // if any of the values in the array are NaN,
    // we make our own explicit maxIndex.
    var maxIndex = 0
    var maxValue = p(maxIndex)
    for (i <- 1 until numClasses) {
      if (p(i) > maxValue) {
        maxIndex = i
        maxValue = p(i)
      }
    }
    maxIndex
  }

}
